import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from scipy.optimize import curve_fit
from scipy.interpolate import CubicSpline

# plt.style.use('./pyDWS/plottingStyle.mplstyle')

from pyDWS.plots import plot_layout

from pyDWS.utils import generate_time_array


def tc_echo_merge(tc_icf, echo_icf, overlapN, overlap_threshold, fitP=False, savePlots=None, saveData=None):   
    """
    Function to merge and blend the two-cell and echo measurements.

    Parameters
    ----------
    tc_icf : dataframe
        two Cell data.
    echo_icf : dataframe
        echo data.
    overlapN : int
        Number of points to use from the echo to fit the second decay to get A and tau_TC. 
    overlap_threshold : float
        Permitted difference between the echo and the two-cell data set in order to merge the two. 
    fitP : bool, optional
        If fitP = True, then A, tauTC, and p are fitted in the second decay. Otherwise, only A, tauTC, and p = 2 is fixed. 
        The default is False.
    savePlots : None or string
        Default None. Otherwise the path where the graphs should be saved. 
    saveData : None or string
        Default None. Otherwise the path where the text files should be saved. 

    Returns
    -------
    All are panda dataframes
    If saveData is set, then these dataframes are saved in txt files. 
    ratioTC_E: columns t & ratio (ratio of two-cell and echo) -> ratioTC_echo.txt
    a_tauTC_p: columns A & tauTC (and p if fitted) second decay fitting parameters -> a_tauTC_p.txt
    tc_icf_overlap: columns t & icf (corrected two-cell data) -> tc_corrected.txt
    echo_icf_corrected: columns t & icf (corrected echo data) -> echo_corrected.txt
    sortedCombined: columns t & icf (merged two-cell and echo without adjusted time spacing) -> tc_echo_merged.txt
    tc_echo_mergeInt: columns t & icf (merged two-cell and echo with adjusted time spacing) -> tc_echo_merged_Interpolation.txt
    
    
    """
    
    # =============================================================================
    # two cell data 
    # =============================================================================
    # interpolate two cell data 
    tc_icf_int = CubicSpline(tc_icf.iloc[:, 0], tc_icf.iloc[:, 1])
    
    # genarate log spaced x data
    startE = np.log10(min(tc_icf.iloc[:, 0]))
    endE = np.log10(round(max(tc_icf.iloc[:, 0])))
    # endE = -1
    xval = np.logspace(startE, endE, 1000)
    # print(xval)

    
    # normalize the interpolation 
    norm_factor = tc_icf.iloc[0, 1]
    def interPolationNorm(x):
        return tc_icf_int(x) / norm_factor
    
    
    tc_icfNorm = tc_icf.iloc[:, 1] / norm_factor
    tc_icf_Int_Norm = interPolationNorm(xval)
    
    
    # plto interpolation 
    fig, ax = plt.subplots(figsize=(6.4,4.8))
    # normalized
    # ax.plot(tc_icf.iloc[:, 0], tc_icf.iloc[:, 1] / norm_factor, 'o', label='two cell measurement')
    # ax.plot(xval, interPolationNorm(xval), '-', label='interpolation', color='red')
    
    ax.plot(tc_icf.iloc[:, 0], tc_icf.iloc[:, 1], 'o', label='two cell measurement')
    ax.plot(xval, tc_icf_int(xval), '-', label='interpolation', color='red')
    
    plt.legend()
    plot_layout(fig, ax, True, False, 'Lag time t (s)', 'ICF')
    if savePlots != None:
        plt.savefig(os.path.join(savePlots, 'twoCell_Interpolation.png'))


    # =============================================================================
    # Echo with overlap
    # =============================================================================
    # plot echo with overlap 
    fig, ax = plt.subplots(figsize=(6.4,4.8))
    ax.plot(echo_icf.iloc[:, 0], echo_icf.loc[:, 'area'] , 'o', label='echo measurement')
    # ax.plot(echo_icf.iloc[:, 0], echo_icf.loc[:, 'ICF'] , 'o', label='echo measurement')
    ax.plot(echo_icf.iloc[:overlapN, 0], echo_icf.iloc[:overlapN, 4], 'o', color='None', markeredgecolor='red')
    

    plt.legend()
    plot_layout(fig, ax, True, False, 'Lag time t (s)', 'ICF')
    if savePlots != None:
        plt.savefig(os.path.join(savePlots, 'echo_Overlap.png'))


    # =============================================================================
    # Mercing two cell echo
    # =============================================================================

    # ratio calculation normalized
    # ratioTC_E = pd.DataFrame({'t': echo_icf.iloc[:overlapN, 0], 'ratio': interPolationNorm(echo_icf.iloc[:overlapN, 0]) / echo_icf.iloc[:overlapN, 4]})
    
    # ratio calculation 
    ratioTC_E = pd.DataFrame({'t': echo_icf.iloc[:overlapN, 0], 'ratio': tc_icf_int(echo_icf.iloc[:overlapN, 0]) / echo_icf.iloc[:overlapN, 4]})
   
       
    
    # fitting of the parameters A and tau_TC or also p
    if fitP == True:
        popt, pcov = curve_fit(_secondDecay_p, ratioTC_E['t'], ratioTC_E['ratio'], p0=(200000, 0.1, 2))
    else:
        # popt, pcov = curve_fit(_secondDecay, ratioTC_E['t'], ratioTC_E['ratio'], p0=(0.1, 0.1))
        popt, pcov = curve_fit(_secondDecay, ratioTC_E['t'], ratioTC_E['ratio'], p0=(200000, 0.1))

    

    # plot of fitting A and tau_TC 
    xratio = np.linspace(echo_icf.iloc[0, 0], echo_icf.iloc[overlapN-1, 0], 100)

    fig, ax = plt.subplots(figsize=(6.4,4.8))
    ax.plot(ratioTC_E['t'], ratioTC_E['ratio'] , 'o')
    if fitP == True:
        ax.plot(xratio, _secondDecay_p(xratio, popt[0], popt[1], popt[2]), '-', color='red')
        
    else:
        ax.plot(xratio, _secondDecay(xratio, popt[0], popt[1]), '-', color='red')
    
    plot_layout(fig, ax, False, False, 'Lag time t (s)', r'$\frac{g_2^{TC}(t)-1}{g_2^{Echo}(t)-1}$')
    if savePlots != None:
        plt.savefig(os.path.join(savePlots, 'second_Decay_fit.png'))


    # Correcting two cell data 
    if fitP == True:
        correctFactor_TC = np.exp((tc_icf.iloc[:, 0] / popt[1])**popt[2])
    else:
        correctFactor_TC = np.exp((tc_icf.iloc[:, 0] / popt[1])**2)
    
    
    # normalized
    # tc_icf_corrected = pd.DataFrame({'t': tc_icf.iloc[:, 0], 'icf': (tc_icf.iloc[:, 1] * correctFactor_TC) / tc_icf.iloc[0, 1]})
    # unnormalized
    tc_icf_corrected = pd.DataFrame({'t': tc_icf.iloc[:, 0], 'icf': ((tc_icf.iloc[:, 1]) * correctFactor_TC)})
    
    
    # removing all inf values
    tc_icf_corrected = tc_icf_corrected[np.isfinite(tc_icf_corrected).all(1)]
    
    
    # calculate interpolation of corrected two cell data
    # tc_icf_corrected_int = CubicSpline(tc_icf_corrected.iloc[:, 0], tc_icf_corrected.iloc[:, 1])

    # Correction echo data
    echo_icf_corrected = pd.DataFrame({'t': echo_icf.iloc[:, 0], 'icf': echo_icf.iloc[:, 4] * popt[0]})
    
    
    echo_icf_corrected = echo_icf_corrected[1:]
    
    echo_icf_corrected_int = CubicSpline(echo_icf_corrected.iloc[:, 0], echo_icf_corrected.iloc[:, 1])
    
    # connecting two cell data and echo data into single dataset
    # echo_overlap = echo_icf_corrected[abs(echo_icf_corrected.loc[:, 'icf'] - tc_icf_corrected_int(echo_icf_corrected.loc[:, 't'])) < overlap_threshold]
    
    echo_icf = echo_icf[1:]
    
    condition = (
    ((tc_icf_corrected['t'] > echo_icf.iloc[0, 0]) & 
     (abs(tc_icf_corrected['icf'] - echo_icf_corrected_int(tc_icf_corrected['t'])) < overlap_threshold))
    | (tc_icf_corrected['t'] <= echo_icf.iloc[0, 0])
    )

    echo_overlap = tc_icf_corrected[condition]

    plotEchoOverlap = echo_overlap[echo_overlap['t'] >= echo_icf.iloc[0, 0]]

    # Plot corrected two-cell & Echo and the overlap region 
    fig, ax = plt.subplots(figsize=(6.4,4.8))
    ax.plot(tc_icf_corrected['t'], tc_icf_corrected['icf'] , 'o', label='two cell')
    ax.plot(echo_icf_corrected['t'], echo_icf_corrected['icf'] , 'o', label='echo')
    ax.plot(plotEchoOverlap['t'], plotEchoOverlap['icf'] , '--', label='overlap')
    plt.legend()
    plot_layout(fig, ax, True, False, 'Lag time t (s)', r'ICF')
    ax.set_ylim(0, 1)
    if savePlots != None:
        plt.savefig(os.path.join(savePlots, 'tc_echo_corrected.png'))


   
    tc_icf_overlap = echo_overlap


    combined = pd.concat([tc_icf_overlap, echo_icf_corrected])

    sortedCombined = combined.sort_values(by=['t'])

    # Plot merged two-cell & echo
    fig, ax = plt.subplots(figsize=(6.4,4.8))
    ax.plot(tc_icf_overlap['t'], tc_icf_overlap['icf'] , 'o', label='two cell')
    ax.plot(echo_icf_corrected['t'], echo_icf_corrected['icf'] , 'o', label='echo')
    ax.plot(sortedCombined['t'], sortedCombined['icf'] , '--', label='merged')
    plt.legend()
    plot_layout(fig, ax, True, False, 'Lag time t (s)', r'ICF')
    if savePlots != None:
        plt.savefig(os.path.join(savePlots, 'tc_echo_merged.png'))
        
        
    # Interpolating tc - echo merge to have same time spacing as for the two-cell measurement
    sortedCombined_InterP = CubicSpline(sortedCombined.iloc[:, 0], sortedCombined.iloc[:, 1])
    tN = generate_time_array()    
    tc_echo_mergeInt = pd.DataFrame({'t':tN, 'icf':sortedCombined_InterP(tN)})
    
    
    fig, ax = plt.subplots(figsize=(6.4,4.8))
    ax.plot(sortedCombined['t'], sortedCombined['icf'] , 'o', label='merged')
    ax.plot(tc_echo_mergeInt['t'], tc_echo_mergeInt['icf'] , '-', label='Interpolation')
    plt.legend()
    plot_layout(fig, ax, True, False, 'Lag time t (s)', r'ICF')
    if savePlots != None:
        plt.savefig(os.path.join(savePlots, 'tc_echo_merged.png'))
    
    # fitting parameters save as a pandas Dataframe
    if fitP == True:
        a_tauTC_p = pd.DataFrame({'A':[popt[0]], 'tauTC':[popt[1]], 'p':[popt[2]]})
    else:
        a_tauTC_p = pd.DataFrame({'A':[popt[0]], 'tauTC':[popt[1]]})
    
    # =============================================================================
    # saving data
    # =============================================================================
    if saveData != None:
        sortedCombined.to_csv(os.path.join(saveData, 'tc_echo_merged.txt'), sep='\t', index=False)
        tc_icf_overlap.to_csv(os.path.join(saveData, 'tc_corrected.txt'), sep='\t', index=False)
        echo_icf_corrected.to_csv(os.path.join(saveData, 'echo_corrected.txt'), sep='\t', index=False)
        tc_echo_mergeInt.to_csv(os.path.join(saveData, 'tc_echo_merged_Interpolation.txt'), sep='\t', index=False)
        ratioTC_E.to_csv(os.path.join(saveData, 'ratioTC_echo.txt'), sep='\t', index=False)
        a_tauTC_p.to_csv(os.path.join(saveData, 'a_tauTC_p.txt'), sep='\t', index=False)
        
    # return sortedCombined.reset_index(drop=True)
    return ratioTC_E, a_tauTC_p, tc_icf_overlap, echo_icf_corrected, sortedCombined.reset_index(drop=True), tc_echo_mergeInt.reset_index(drop=True)
    





# =============================================================================
# Private functions
# =============================================================================
# second decay fit function 
def _secondDecay(t, A, tauTC):
    """
    Second deay function from which A and tauTC can be extracted
    Parameters
    ----------
    t : Dataframe
        lag time t .
    A : float
        Prefactor.
    tauTC : float
        decay time.

    Returns
    -------
    Second deay function.

    """
    return A * np.exp(-(t / tauTC)**2)


def _secondDecay_p(t, A, tauTC, p):
    """
    Second deay function from which A and tauTC can be extracted.
    Parameters
    ----------
    t : TYPE
        DESCRIPTION.
    A : TYPE
        DESCRIPTION.
    tauTC : TYPE
        DESCRIPTION.
    p : float, optional
        If the fluctuation become noticable, p can be fitted (adjustable parameter). The default is 2.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    return A * np.exp(-(t / tauTC)**p)










import numpy as np


def generate_EchoLayout(tstep, f, save=False):
    """
    Generating Echo layout to perform linear correlation measurement with the LS Instruments Correlator
    
    So the rule is:
    Start with 1–8
    Then generate blocks of 7 numbers with increasing step sizes (2, 3, 4, ..., 11)

    Parameters
    ----------
    tstep : float
        Linear Tau Sampling time.
    f : float
        frequency.
    save : bol or string
        If you want to save it to a text file (default False).
        If you want to save it, you have to define the path. 

    Returns
    -------
    None.

    """
    # Initial time vector
    t = list(range(1, 7))  
    t0 = 1/f
    
    # generating peak distribution 
    all_peak = []
    
    # Subsequent blocks
    current = 1
    for step in range(1, 11):  # Steps from 2 to 11
        for _ in range(7): # Repeat this loop 7 times, but I don’t care about the loop variable
            all_peak.append(current)
            current += step
    all_peak.append(current) # adding one last aditionally 
    print('Peak distribution: ', all_peak)
    
          
    

    # Expand time points around each peak
    for peak in all_peak:
        sumit = int(np.floor(peak * t0 / tstep))
        t1 = list(range(sumit - 6, sumit + 7))
        t.extend(t1)

    # Convert to numpy array and reshape as a column vector
    t = np.array(t).reshape(-1, 1)
    
    if save != False:
        with open(save, 'w') as file:
            for val in t:
                file.write(f"{val[0]}\n")

    return t






# generate time scaling as for multitau measurement
def generate_time_array(start=1.25e-8, stop=12.6, initial_step=1.25e-8, steps_per_block=8, step_growth=2):
    """
    Generates a time array that starts with small uniform steps and increases the step size every block.
    
    Parameters:
    - start: Initial time value
    - stop: Maximum time value to reach
    - initial_step: Starting step size
    - steps_per_block: Number of steps before step size increases
    - step_growth: Multiplicative factor to grow step size
    
    Returns:
    - np.ndarray of time values
    """
    time_values = []
    current = start
    step = initial_step
    
    for i in range(16):
        time_values.append(current)
        current += step
    
    step *= step_growth
    # print(current)
    
    while current < stop:
        for _ in range(steps_per_block):
            if current >= stop:
                break
            time_values.append(current)
            current += step
        step *= step_growth

    time_values.insert(0, 1e-8)
    return np.array(time_values)
















import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os

from scipy.signal import butter, lfilter
from scipy.interpolate import CubicSpline
from scipy.stats import pearsonr
import statsmodels.api as sm
from cmath import sqrt

# plt.style.use('./pyDWS/plottingStyle.mplstyle')

from pyDWS.plots import plot_layout




def microrheology(msdData, temperatur, freq_range, gau, radi, zPath=None, savePlots=None):
    """
    Calculation of the complex modulus based on:
    T. G. Mason, Estimating the viscoelastic moduli of complex fluids using the generalized stokes–einstein equation, Rheologica acta 39, 371 (2000).
    

    Curve Smoothing with lowess:
    Seabold, Skipper, and Josef Perktold. “statsmodels: Econometric and statistical modeling with python.” Proceedings of the 9th Python in Science Conference. 2010.
    

    Parameters
    ----------
    msdData : pandas Dataframe
        First column contains t and second column msd.
    temperatur : float
        temperatur at which measurement was performed.
    freq_range : list
        frequency range of the moduli.
    gau : tuple
        If len(gau) == 2, then a low-pass filter is applied to the derivative values. The first value corresponds to the order of the filter, and the second to the critical frequency.
        Else there is no filtering. 
    radi : float
        radius of the particles in the sample.
    zPath : string, optional
        Path to save the microrheology data. The default is None.

    Returns
    -------
    pdZ : pandas dataframe
    
    if filtering is performed:
        columns:
            ome: angular frequency
            zstar: Z*
            zpr: Z'
            zppr: Z''
            zstarFi: Z* with applied filter
            zprFi: Z' with applied filter
            zpprFi: Z'' with applied filter
            zprS: Z' with applied filter and with applied smoothing
            zpprS: Z'' with applied filter and with applied smoothing
            zC: complex Z* 
            
    else:
        columns:
            ome: angular frequency
            zstar: Z*
            zpr: Z'
            zppr: Z''
            zC: complex Z* 

    """
    
    
    if freq_range is not None:
        low = freq_range[0]
        high = freq_range[1]
        cmsd = msdData[(msdData.iloc[:, 0] >= 1 / high) & (msdData.iloc[:, 0] <= 1 / low)]


    cmsd = cmsd.to_numpy()
    # print(cmsd)
    
    ox = cmsd[:, 0]
    oy = cmsd[:, 1]
     
    
    x = np.log(ox)
    y = np.log(oy)
    
    dx = np.diff(x)
    dy = np.diff(y)
    
    slp = dy / dx
    xc = (x[:-1] + x[1:]) / 2
    
    # interpolation of the derivatives
    interp_func = CubicSpline(xc, slp, extrapolate=True)
    deriv = interp_func(x)
   
    
    
    ome = 1. / ox # angular frequency
    kB = 1.38064852e-23
    
    # smoothing 
    if isinstance(gau, (list, tuple)) and len(gau) == 2:
        print('Smoothing using Butterworth Lowpass Filter...')
        derivFi = butter_low_pass(deriv, gau[0], gau[1])
    
        
        # calculate  Z* Z' and Z'' with filtered derivatives
        gamaFi = 0.457 * (1 + derivFi)**2 - 1.36 * (1 + derivFi) + 1.90
        zstarFi = kB * temperatur / (np.pi * radi * oy * gamaFi)
        angFi = np.pi * derivFi / 2
        zprFi = zstarFi * np.cos(angFi)
        zpprFi = zstarFi * np.sin(angFi)
    
    
        # calculate  Z* Z' and Z'' without filtering
        gama = 0.457 * (1 + deriv)**2 - 1.36 * (1 + deriv) + 1.90
        zstar = kB * temperatur / (np.pi * radi * oy * gama)
        ang = np.pi * deriv / 2
        zpr = zstar * np.cos(ang)
        zppr = zstar * np.sin(ang)
    
        
        # Plot MSD
        fig, ax = plt.subplots(figsize=(6.4,4.8))
        ax.plot(ox, oy, 'ob')
        plot_layout(fig, ax, True, True, r'Lag time t (s)', r'MSD')
        
        
        
        # Plot Z* Z' and Z'' without filtering and filtering
        fig, ax = plt.subplots(figsize=(6.4,4.8))
        ax.plot(ome, zstar, '-k', linewidth=3, label=r"$ G* (\omega) $")
        ax.plot(ome, zpr, 'ob', label=r"$ G' (\omega) $")
        ax.plot(ome, zppr, 'or', label=r"$ G'' (\omega) $")
        
        ax.plot(ome, zstarFi, '-', linewidth=3, label=r"$ G* (\omega) $ filtered", color='gray')
        ax.plot(ome, zprFi, '--', linewidth=2, label=r"$ G' (\omega) $ filtered", color='gray')
        ax.plot(ome, zpprFi, '-.', linewidth=2, label=r"$ G'' (\omega) $ filtered", color='gray')
        
        ax.legend()
        plot_layout(fig, ax, True, True, r'Angular frequency $\omega$ (rad/s)', r'Modulus (Pa)')
        if savePlots != None:
            plt.savefig(os.path.join(savePlots, 'Z_filter.png'))
    
        
        # smoothing Z' & Z''
        zpr_smooth = sm.nonparametric.lowess(zprFi, ome, frac=0.1, it=0, delta=0.0, is_sorted=False)
        zppr_smooth = sm.nonparametric.lowess(zpprFi, ome, frac=0.1, it=0, delta=0.0, is_sorted=False)
        
        
        # Plot Z* Z' and Z'' but smoothed
        fig, ax = plt.subplots(figsize=(6.4,4.8))
        ax.plot(ome, zstarFi, '-k', linewidth=3, label=r"$ G* (\omega) $")
        ax.plot(ome, zprFi, 'ob', label=r"$ G' (\omega) $")
        ax.plot(ome, zpprFi, 'or', label=r"$ G'' (\omega) $")
        
        ax.plot(zpr_smooth[:,0], zpr_smooth[:,1], '-k', linewidth=2, label=r"$ G' (\omega) $ smooth")
        ax.plot(zppr_smooth[:,0], zppr_smooth[:,1], '-k', linewidth=2, label=r"$ G'' (\omega) $ smooth")
        
        ax.legend()
        plot_layout(fig, ax, True, True, r'Angular frequency $\omega$ (rad/s)', r'Modulus (Pa)')
        if savePlots != None:
            plt.savefig(os.path.join(savePlots, 'Z_filterSmooth.png'))
    
        # saving microrheology data
        pdZ = pd.DataFrame({'ome': ome, 'zstar': zstar, 'zpr': zpr, 'zppr': zppr, 'zstarFi': zstarFi, 'zprFi': zprFi, 'zpprFi': zpprFi, 'zprS': np.flip(zpr_smooth[:,1]), 'zpprS': np.flip(zppr_smooth[:,1])})
        pdZ['zC'] = pdZ.loc[:, 'zprS'] + (pdZ.loc[:, 'zpprS'] * 1j)
        
        
        if zPath != None:
            pdZ.to_csv(os.path.join(zPath, 'zmicrorheology-filtered.txt'), sep='\t', index=False)
            
        
        return pdZ

    
    else:
        # calculate  Z* Z' and Z''
        gama = 0.457 * (1 + deriv)**2 - 1.36 * (1 + deriv) + 1.90
        zstar = kB * temperatur / (np.pi * radi * oy * gama)
        ang = np.pi * deriv / 2
        zpr = zstar * np.cos(ang)
        zppr = zstar * np.sin(ang)
    
    
        # Plot MSD
        fig, ax = plt.subplots(figsize=(6.4,4.8))
        ax.plot(ox, oy, 'ob')
        plot_layout(fig, ax, True, True, r'Lag time t (s)', r'MSD')
        
        
        # Plot Z* Z' and Z'' with filtering 
        fig, ax = plt.subplots(figsize=(6.4,4.8))
        ax.plot(ome, zstar, '-k', linewidth=3, label=r"$ G* (\omega) $")
        ax.plot(ome, zpr, 'ob', label=r"$ G' (\omega) $")
        ax.plot(ome, zppr, 'or', label=r"$ G'' (\omega) $")
        
        ax.legend()
        plot_layout(fig, ax, True, True, r'Angular frequency $\omega$ (rad/s)', r'Modulus (Pa)')
        if savePlots != None:
            plt.savefig(os.path.join(savePlots, 'Z_nofilter.png'))
        
        
        # saving microrheology data
        pdZ = pd.DataFrame({'ome': ome, 'zstar': zstar, 'zpr': zpr, 'zppr': zppr})
        pdZ['zC'] = pdZ.loc[:, 'zpr'] + (pdZ.loc[:, 'zppr'] * 1j)
        
        
        if zPath != None:
            pdZ.to_csv(os.path.join(zPath, 'zmicrorheology.txt'), sep='\t', index=False)
    
             
        return pdZ
    


def butter_low_pass(data_in, order, smooth):
    """
    Low-pass Butterworth filter

    Parameters
    ----------
    data_in : numpy array
        Input data (1D NumPy array).
    order : int
        Filter order.
    smooth : float
        Normalized cutoff frequency (0 < smooth < 1).

    Returns
    -------
    result : numpy array
        Filtered data (same shape as input).

    """
    
    b, a = butter(order, smooth, btype='lowpass')

    # Extend the data at both ends to minimize edge effects
    data_in_ext = np.concatenate([np.full_like(data_in, data_in[0]), data_in, np.full_like(data_in, data_in[-1])])
    

    # Apply filter

    data_out = lfilter(b, a, data_in_ext)
    # print(data_out)

    # Try different shifts and compute correlation
    max_corr = -1
    best_shift = 0
    for i in range(1, 101):
        temp = np.roll(data_in_ext, i)
        corr, _ = pearsonr(temp, data_out)
        if corr > max_corr:
            max_corr = corr
            best_shift = i

    # Align the output
    data_out = data_out[best_shift:]

    # Extract central portion (original length)
    third = len(data_in_ext) // 3
    result = data_out[third:2*third]

    return result





def inertia_correction(omega, zcomplex, radi, rho_b, rho, ranges=[[1e-1, 5*1e6], [1, 1e5]], gPath=None, savePlots=None):    
    """
    Inertia correction based on:
    Domínguez-García, F. Cardinaux, E. Bertseva, L. Forró, F. Scheffold, and S. Jeney, Accounting for inertia effects to access the high-frequency microrheology of viscoelastic fluids, Physical Review E 90, 060301 (2014).

    Parameters
    ----------
    omega : numpy array
        angular frequency.
    zcomplex : numpy array
        complex modulus.
    radi : float
        radius of the particles in the sample.
    rho_b : float
        density of the particles (kg/m^3).
    rho : float
        medium density (kg/m^3).
    ranges : 2d list
        first row x axis range and second row y axis range
    gPath : string, optional
        Path to save the microrheology data. The default is None.
    savePlots : None or string
        Default None. Otherwise the path where the graphs should be saved. 

    Returns
    -------
    g_corrected : pandas dataframe
        columns:
            ome: angular frequency
            gStar: G*
            gStarABS: |G*|
            gpr: G'
            gppr: G''

    """
    
    
    # Compute m* (effective mass)
    mstar = ((4 * np.pi * radi**3 * rho_b) / 3) + ((2 * np.pi * rho * radi**3) / 3)

    # performing inertia Correction 
    term2 = (mstar * omega**2) / (6 * np.pi * radi)
    inner_term = rho**2 - (((2 * rho) / (3 * np.pi * radi**3)) * ((((6 * np.pi * radi) / omega**2) * zcomplex) + mstar))

    term3 = []
    for i in range(len(inner_term)):
        term3.append((radi**2 * omega[i]**2) / 2 * (sqrt(inner_term[i])-rho))


    gStar = (zcomplex + term2 + term3)
    gStarABS = np.abs(gStar)

    gpr = gStar.real
    gppr = gStar.imag
    
    
    # Preparing data for the plot in the correct range
    ranges = np.array(ranges, dtype=float)
    mask = np.logical_or.reduce([(omega >= lo) & (omega <= hi) for lo, hi in ranges])
    
    omegaP = omega[mask]
    gStarABSP = gStarABS[mask]
    gprP = gpr[mask]
    gpprP = gppr[mask]
    
        
    
    # Plot G* G' and G'' with filtering 
    fig, ax = plt.subplots(figsize=(6.4,4.8))
    ax.plot(omegaP, gStarABSP, '-k', linewidth=3, label=r"$ G* (\omega) $")
    ax.plot(omegaP, gprP, 'ob', label=r"$ G' (\omega) $")
    ax.plot(omegaP, gpprP, 'or', label=r"$ G'' (\omega) $")
    
    ax.legend()
    plot_layout(fig, ax, True, True, r'Angular frequency $\omega$ (rad/s)', r'Modulus (Pa)')
    if savePlots != None:
        plt.savefig(os.path.join(savePlots, 'G_inertia_correction.png'))


    g_corrected = pd.DataFrame({'ome': omega, 'gStar': gStar, 'gStarABS': gStarABS, 'gpr': gpr, 'gppr': gppr})

    if gPath != None:
        g_corrected.to_csv(os.path.join(gPath, 'gmicrorheology-corrected.txt'), sep='\t', index=False)


    return g_corrected










